#!/bin/bash

srcdir="$(builtin cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
PATH="$srcdir/scripts:$PATH"

source "$srcdir"/helpers/log.bash
source "$srcdir"/helpers/utils.bash
source "$srcdir"/helpers/output.bash
source "$srcdir"/helpers/query.bash

############## RPKG ENV VARS ##############

declare -x INPUT_PATH
declare -x INPUT_DIR_PATH
declare -x INPUT_GIT_DIRTY
declare -x INPUT_GIT_REVISION
declare -x VERSION_BUMP
declare -x RELEASE_BUMP
declare -x VERBOSE
declare -x OUTDIR

############## PERSISTENT STATE ##############

declare -x GIT_VERSION
declare -x GIT_LEGACY
declare -x GIT_ROOT
declare -x GIT_BRANCH
declare -x GIT_HEAD
declare -x GIT_HEAD_SHORT
declare -x GIT_ORIGIN_URL

GIT_VERSION="${GIT_VERSION:-$(git --version)}"
GIT_LEGACY="${GIT_LEGACY:-$(echo "$GIT_VERSION" | grep 'git version 1')}"
GIT_ROOT="${GIT_ROOT:-$(git rev-parse --show-toplevel 2> /dev/null)}"
GIT_BRANCH="${GIT_BRANCH:-$(basename "$(git -C "$GIT_ROOT" symbolic-ref HEAD 2> /dev/null)")}"
GIT_HEAD="${GIT_HEAD:-${INPUT_GIT_REVISION:-$(git -C "$GIT_ROOT" rev-parse --revs-only "$GIT_BRANCH" 2> /dev/null)}}"
GIT_HEAD_SHORT="${GIT_HEAD_SHORT:-$(expr substr "$GIT_HEAD" 1 7)}"
GIT_ORIGIN_URL="${GIT_ORIGIN_URL:-$(git -C "$GIT_ROOT" config --get remote.origin.url 2> /dev/null)}"

############## HELPERS ##############

function git_check_path {
    if [ -z "$1" ]; then
        log_error "path cannot be empty."
        return 1
    fi

    if ! [ -d "$1" ]; then
        log_error "path is not a directory."
        return 1
    fi

    abs_path="$(builtin cd "$1" && pwd -P)"

    if ! [[ "$abs_path" =~ ^$GIT_ROOT.*$ ]]; then
        log_error "path outside of $GIT_ROOT."
        return 1
    fi
}

function git_check_repo {
    if [ -z "$GIT_ROOT" ]; then
        log_error "Not a Git repository."
        return 1
    fi
}

function git_filter_tags {
    grep -E "^$1-[^-]+-[^-]+$"
}

function git_prefix_to_name_prefix {
    sed -e 's|//*|-|g' -e 's|\([^-]\)$|-\1|' -e 's|^-||'
}

function git_prefix_to_name_suffix {
    sed -e 's|//*|-|g' -e 's|^\([^-]\)|-\1|' -e 's|-$||'
}

function git_formatted_changelog_record {
    declare data= name= tag= header_locale= header_date_format= body_wrap= "$@"

    verrel="$(echo "$tag" | sed -E -n "s/^$name-([^-]+)(-[^-]+)$/\1\2/p")"

    if [ -n "$GIT_LEGACY" ]; then
        if [ -n "$header_date_format" ]; then
             realdate="$(git -C "$GIT_ROOT" for-each-ref "refs/tags/$tag" --format '%(taggerdate:iso8601)')"
             header_date="$(date --date "$realdate" +"$header_date_format")"
        fi
        header="$(LC_ALL="$header_locale" git -C "$GIT_ROOT" for-each-ref "refs/tags/$tag" \
            --format="* $header_date %(taggername) %(taggeremail)") $verrel"
    else
        if [ -n "$header_date_format" ]; then
            header_date_format_suffix=":format:$header_date_format"
        else
            header_date_format_suffix=
        fi
        header="$(LC_ALL="$header_locale" git -C "$GIT_ROOT" tag -l "$tag" \
            --format="* %(taggerdate$header_date_format_suffix) %(taggername) %(taggeremail)") $verrel"
    fi

    formatted_data="$(echo "$data" | fold -cs --width="$body_wrap" | sed -E -e 's/(\s+)%([^%])/\1%%\2/g' -e 's/\s*$//')"
    echo -n "${header}"$'\n'"${formatted_data}"
}

function git_bumped_version {
    declare lead=0 follow= "$@"

    if [ -z "$follow" ] && [ -n "$lead" ] && [ "$lead" != 0 ]; then
        echo -n "${lead}.0"
        return
    fi

    if echo "$follow" | grep -q '[^0-9]'; then
        log_error "$follow is not a number. Please, bump version manually."
        return 1
    fi

    bumped_follow=$(( $follow + 1 ))
    echo -n "${lead:+$lead.}${bumped_follow}"
}

function git_bumped_release {
    declare pivot= "$@"

    if echo "$pivot" | grep -q '[^0-9]'; then
        log_error "$pivot is not a number. Please, bump release manually."
        return 1
    fi

    bumped_pivot=$(( $pivot + 1 ))
    echo -n "$bumped_pivot"
}

############## QUERIES ##############

function git_branch_remote_url {
    if [ -n "$GIT_LEGACY" ]; then
        git -C "$GIT_ROOT" remote -v | grep '(fetch)' | sed -E -e 's/(.+)\s+(.+)\s+(.+)\n*/\2/'
    else
        git -C "$GIT_ROOT" remote get-url "$(git -C "$GIT_ROOT" config branch."$GIT_BRANCH".remote)" 2> /dev/null
    fi
}

function git_revlist_since {
    git -C "$GIT_ROOT" rev-list "$1..$GIT_HEAD"
}

function git_revlist {
    git -C "$GIT_ROOT" rev-list "$GIT_HEAD"
}

function git_version_dirty_appendix {
    dirty_appendix=

    if [ -n "$INPUT_GIT_DIRTY" ] || [ "$GIT_HEAD" != "$(git rev-parse HEAD 2> /dev/null)" ]; then
        dirty_appendix=".dirty"
    fi

    query git_latest_ctime
    latest_ctime="${QUERY[git_latest_ctime]%$'\n'}"
    if [ "$latest_ctime" -ne 0 ]; then
        dirty_appendix=".dirty.$(utils_encode_decimal "$latest_ctime")"
    fi

    if [ -n "$dirty_appendix" ]; then
        echo "$dirty_appendix"
    fi
}

function git_version_appendix {
    if [ -n "$1" ]; then
        query git_revlist_since "$1"
        commit_count="$(printf "%s" "${QUERY["git_revlist_since $1"]}" | wc -l)"
    else
        query git_revlist 2> /dev/null
        commit_count="$(printf "%s" "${QUERY[git_revlist]}" | wc -l || printf 0)"
    fi

    if [ "$commit_count" -eq 0 ]; then
        commit_count_appendix=
    else
        commit_count_appendix=".git.$commit_count.$GIT_HEAD_SHORT"
    fi

    query git_version_dirty_appendix
    version_appendix="${commit_count_appendix}${QUERY[git_version_dirty_appendix]%$'\n'}"

    if [ -n "$version_appendix" ]; then
        echo "$version_appendix"
    fi
}

function git_merged_tags {
    if [ -n "$GIT_LEGACY" ]; then
        git -C "$GIT_ROOT" for-each-ref --sort=-taggerdate refs/tags | cut -f2 |
            while read -r tag; do
                if [ -z "$(basename "$tag" | git_filter_tags "$1")" ]; then
                    continue
                fi
                if [ "$(git -C "$GIT_ROOT" merge-base "$tag" "$GIT_HEAD")" = "$(git -C "$GIT_ROOT" rev-parse "$tag"^{commit})" ]; then
                    basename "$tag"
                fi
            done
    else
        git -C "$GIT_ROOT" tag --list --sort=-taggerdate "$1*" --merged 2> /dev/null | git_filter_tags "$1"
    fi
}

function git_url_path_name {
    query git_remote_url
    basename="$(basename "${QUERY[git_remote_url]%$'\n'}")"

    if [ -z "$basename" ]; then
        return
    fi

    query git_prefix "$1"
    suffix="$(printf "%s" "${QUERY["git_prefix $1"]}" | git_prefix_to_name_suffix)"

    echo "${basename%.git}${suffix}"
}

function git_prefix {
    git -C "${1:-$GIT_ROOT}" rev-parse --show-prefix
}

function git_status {
    git -C "$GIT_ROOT" status --porcelain -uno "${1:-$GIT_ROOT}"
}

function git_latest_ctime {
    latest_ctime=0
    query git_status
    while read -r f; do
        if [ -e "$GIT_ROOT/$f" ]; then
            ctime=$(stat --format='%Z' "$GIT_ROOT/$f")
            if [ "$ctime" -gt "$latest_ctime" ]; then
                latest_ctime="$ctime"
            fi
        fi
    done < <(printf "${QUERY[git_status]}" | cut -c 4- | sed -E -e 's|(.*) -> (.*)|\1\n\2|')
    echo "$latest_ctime"
}

function git_remote_url {
    query git_branch_remote_url
    url="${QUERY[git_branch_remote_url]%$'\n'}"

    if [ -z "$url" ]; then
        url="$GIT_ORIGIN_URL"
    fi

    if [ -n "$url" ]; then
        echo "$url"
    fi
}

############## BASE FUNCTIONS ##############

function git_name {
    declare name= prepend= append= "$@"

    if [ -n "$name" ]; then
        output "${prepend}${name}${append}"
        return
    fi

    git_check_repo || return 1

    query git_remote_url
    url="${QUERY[git_remote_url]%$'\n'}"

    if [ -z "$url" ]; then
        log_error "Could not get remote URL."
        return 1
    fi

    url_last_part="$(basename "$url")"
    name="${url_last_part%.git}"

    output "${prepend}${name}${append}"
}

function git_version {
    declare name="${OUTPUT[git_name]}" lead=0 follow= "$@"

    git_check_repo || return 1

    if [ -z "$name" ]; then
        log_error "name cannot be empty."
        return 1
    fi

    if echo "$lead.$follow" | grep -q '-'; then
        log_error "lead and follow cannot contain dashes."
        return 1
    fi

    if echo "$follow" | grep -q '\.'; then
        log_error "follow cannot contain dots."
        return 1
    fi

    query git_merged_tags "$name"
    latest_tag="$(printf "%s" "${QUERY["git_merged_tags $name"]}" | head -1)"
    latest_tag_version="$(echo "$latest_tag" | sed -E -n "s/^${name}-(${lead}[^-]+)-[^-]+$/\1/p")"

    if [ -z "$lead" ]; then
        lead="$(echo "$latest_tag_version" | sed -E -n "s/^(.*)\.[^.]*$/\1/p")"
    fi

    if [ -n "$follow" ]; then
        if [ -n "$VERSION_BUMP" ]; then
            output "${lead:+$lead.}${follow}"
        else
            query git_version_appendix "$latest_tag"
            output "${lead:+$lead.}${follow}${QUERY["git_version_appendix $latest_tag"]%$'\n'}"
        fi
        return
    fi

    follow="$(echo "$latest_tag_version" | sed -E -n "s/^${lead}\.?([^.]*)$/\1/p")"

    if [ -n "$VERSION_BUMP" ]; then
        version="$(git_bumped_version lead="$lead" follow="$follow")"

        retcode=$?
        if [ $retcode -ne 0 ]; then
            return $retcode
        fi

        output "$version"
        return
    fi

    query git_version_appendix "$latest_tag"
    output "${lead:+$lead.}${follow:-0}${QUERY["git_version_appendix $latest_tag"]%$'\n'}"
}

function git_release {
    declare name="${OUTPUT[git_name]}" pivot= "$@"

    git_check_repo || return 1

    if [ -z "$name" ]; then
        log_error "name cannot be empty."
        return 1
    fi

    if echo "$pivot" | grep -q '-'; then
        log_error "pivot cannot contain dashes."
        return 1
    fi

    query git_merged_tags "$name"
    latest_tag="$(printf "%s" "${QUERY["git_merged_tags $name"]}" | head -1)"

    if [ -n "$pivot" ]; then
        if [ -n "$RELEASE_BUMP" ]; then
            output "${pivot}"
        else
            query git_version_appendix "$latest_tag"
            output "${pivot}${QUERY["git_version_appendix $latest_tag"]%$'\n'}"
        fi
        return
    fi

    pivot="$(echo "$latest_tag" | sed -E -n "s/^$name-[^-]+-([^-]+)$/\1/p")"

    if [ -n "$RELEASE_BUMP" ]; then
        pivot="$(git_bumped_release pivot="$pivot")"

        retcode=$?
        if [ $retcode -ne 0 ]; then
            return $retcode
        fi

        output "$pivot"
        return
    fi

    query git_version_appendix "$latest_tag"
    output "${pivot:-0}${QUERY["git_version_appendix $latest_tag"]%$'\n'}"
}

function git_changelog {
    declare name="${OUTPUT[git_name]}" since_tag= until_tag= header_locale=POSIX header_date_format="%a %b %d %Y" body_wrap=80 "$@"

    git_check_repo || return 1

    if [ -z "$name" ]; then
        log_error "name cannot be empty."
        return 1
    fi

    changelog=
    until_tag_hit=
    query git_merged_tags "$name"
    while read -r tag; do
        if [ -n "$until_tag" ] && [ -z "$until_tag_hit" ] && [ "$until_tag" != "$tag" ]; then
            continue
        fi
        until_tag_hit=1

        data="$(git -C "$GIT_ROOT" for-each-ref "refs/tags/$tag" --format='%(body)')"
        if [ -z "$data" ]; then
            data="(none)"
        fi

        formatted_record="$(git_formatted_changelog_record \
            data="$data" name="$name" tag="$tag" header_locale="$header_locale" header_date_format="$header_date_format" body_wrap="$body_wrap")"

        if [ -n "$changelog" ]; then
            changelog="$changelog"$'\n\n'
        fi

        changelog="${changelog}${formatted_record}"

        if [ "$since_tag" = "$tag" ]; then
            break
        fi
    done < <(printf "%s" "${QUERY["git_merged_tags $name"]}")

    output "$changelog"
}

function git_vcs {
    declare path= "$@"

    git_check_repo || return 1
    git_check_path "$path" || return 1

    query git_remote_url
    url="${QUERY[git_remote_url]%$'\n'}"

    if [ -z "$url" ]; then
        log_error "Could not get remote URL."
        return 1
    fi

    if [ -n "$GIT_HEAD" ]; then
        query git_prefix "$path"
        tree_suffix="#$GIT_HEAD:${QUERY["git_prefix $path"]%$'\n'}"
    else
        tree_suffix=
    fi

    output "git+${url}${tree_suffix%/}"
}

function git_pack {
    declare path= dir_name= source_name= "$@"

    query git_version_dirty_appendix
    if [ -z "${QUERY[git_version_dirty_appendix]}" ]; then
        git_archive_output="$(git_archive "$@")"
        git_archive_retval=$?

        output "$git_archive_output"
        return "$git_archive_retval"
    fi

    git_check_repo || return 1
    git_check_path "$path" || return 1

    if [ -z "$dir_name" ]; then
        query git_url_path_name "$path"
        dir_name="${QUERY["git_url_path_name $path"]%$'\n'}"
    fi

    if [ -z "$dir_name" ]; then
        log_error "Could not derive dir_name from remote URL and path."
        return 1
    fi

    if [ -z "$source_name" ]; then
        dirty_flag=

        query git_status "$path"
        if [ -n "${QUERY["git_status $path"]}" ]; then
            dirty_flag="-dirty"
        fi

        source_name="${dir_name}-${GIT_HEAD_SHORT:-$GIT_BRANCH}${dirty_flag}.tar.gz"
    fi

    if [ -z "$OUTDIR" ]; then
        log_debug "OUTDIR is not set. No action taken."
        output "$source_name"
        return
    fi

    if echo "$GIT_VERSION" | grep -qE '1\.[0-7].*'; then
        log_error "This function is supported up from git 1.8"
        return 1
    fi

    exclude_tmpfile="$(mktemp)"
    git -C "$GIT_ROOT" status --porcelain --ignored -unormal "$path" |
        sed -E -n -e "s+/$++" -e "s+(^!! |^\?\? )+$GIT_ROOT/+p" > "$exclude_tmpfile"

    log_info "packing path $path"
    cmd_stdout="$(pack_sources --exclude-from "$exclude_tmpfile" "$path" "$dir_name" "$OUTDIR/$source_name")"
    cmd_retcode=$?

    log_debug "$cmd_stdout"
    rm "$exclude_tmpfile"

    if [ $cmd_retcode -eq 0 ]; then
        log_info "Wrote: $OUTDIR/$source_name"
    else
        log_error "Error during source creation."
        return $cmd_retcode
    fi

    output "$source_name"
}

function git_archive {
    declare path= dir_name= source_name= "$@"

    git_check_repo || return 1
    git_check_path "$path" || return 1

    if [ -z "$GIT_HEAD" ]; then
        log_error "No commits yet. Content cannot be archived."
        return 1
    fi

    if [ -z "$dir_name" ];  then
        query git_url_path_name "$path"
        dir_name="${QUERY["git_url_path_name $path"]%$'\n'}"
    fi

    if [ -z "$dir_name" ]; then
        log_error "Could not derive dir_name from remote URL and path."
        return 1
    fi

    if [ -z "$source_name" ]; then
        source_name="${dir_name}-${GIT_HEAD_SHORT}.tar.gz"
    fi

    if [ -z "$OUTDIR" ]; then
        log_debug "OUTDIR is not set. No action taken."
        output "$source_name"
        return
    fi

    log_info "archiving $path:"$'\n'"$(git log -1 "$GIT_HEAD" --decorate)"

    cmd_stdout="$(git -C "$path" archive --prefix="$dir_name/" -o "$OUTDIR/$source_name" "$GIT_HEAD")"
    cmd_retcode=$?

    log_debug "$cmd_stdout"

    if [ $cmd_retcode -eq 0 ]; then
        log_info "Wrote: $OUTDIR/$source_name"
    else
        log_error "Error during source creation."
        return $cmd_retcode
    fi

    output "$source_name"
}

function git_setup_macro {
    declare path= dir_name= "$@"

    git_check_repo || return 1

    if [ -z "$dir_name" ]; then
        git_check_path "$path" || return 1

        query git_url_path_name "$path"
        dir_name="${QUERY["git_url_path_name $path"]%$'\n'}"
    fi

    if [ -z "$dir_name" ]; then
        log_error "Could not derive dir_name from remote URL and path."
        return 1
    fi

    output "%setup -q -n $dir_name"
}

############## REPO LEVEL FUNCTIONS ##############

function git_repo_name {
    git_name "$@"
}

function git_repo_version {
    git_version name="$(git_repo_name)" "$@"
}

function git_repo_release {
    git_release name="$(git_repo_name)" "$@"
}

function git_repo_changelog {
    git_changelog name="$(git_repo_name)" "$@"
}

function git_repo_vcs {
    git_vcs path="$GIT_ROOT" "$@"
}

function git_repo_pack {
    git_pack path="$GIT_ROOT" "$@"
}

function git_repo_archive {
    git_archive path="$GIT_ROOT" "$@"
}

function git_repo_setup_macro {
    git_setup_macro path="$GIT_ROOT" "$@"
}

############## DIR LEVEL FUNCTIONS ##############

function git_dir_name {
    git_name append="$(git_prefix "$INPUT_DIR_PATH" | git_prefix_to_name_suffix)" "$@"
}

function git_dir_version {
    git_version name="$(git_dir_name)" "$@"
}

function git_dir_release {
    git_release name="$(git_dir_name)" "$@"
}

function git_dir_changelog {
    git_changelog name="$(git_dir_name)" "$@"
}

function git_dir_vcs {
    git_vcs path="$INPUT_DIR_PATH" "$@"
}

function git_dir_pack {
    git_pack path="$INPUT_DIR_PATH" "$@"
}

function git_dir_archive {
    git_archive path="$INPUT_DIR_PATH" "$@"
}

function git_dir_setup_macro {
    git_setup_macro path="$INPUT_DIR_PATH" "$@"
}

############## CURRENT WORKING DIR FUNCTIONS ##############

function git_cwd_name {
    git_name append="$(git_prefix "$PWD" | git_prefix_to_name_suffix)" "$@"
}

function git_cwd_version {
    git_version name="$(git_cwd_name)" "$@"
}

function git_cwd_release {
    git_release name="$(git_cwd_name)" "$@"
}

function git_cwd_changelog {
    git_changelog name="$(git_cwd_name)" "$@"
}

function git_cwd_vcs {
    git_vcs path="$PWD" "$@"
}

function git_cwd_pack {
    git_pack path="$PWD" "$@"
}

function git_cwd_archive {
    git_archive path="$PWD" "$@"
}

function git_cwd_setup_macro {
    git_setup_macro path="$PWD" "$@"
}
