# Print a man page from the help texts.

# Copyright (C) 2018 Red Hat Inc.
# Author(s): clime <clime@redhat.com>
#
# Original work by:
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.


import os
import sys
import datetime


# We could substitute the "" in .TH with the rpkg version if we knew it
man_header = """\
.\\" man page for rpkg
.TH rpkg 1 "%(today)s" "" "rpm\\-packager"
.SH "NAME"
rpkg \\- RPM packaging utility
.SH "SYNOPSIS"
.B "rpkg"
[
.I global_options
]
.I "command"
[
.I command_options
]
[
.I command_arguments
]
.br
.B "rpkg"
.B "help"
.br
.B "rpkg"
.I "command"
.B "\\-\\-help"
.SH "DESCRIPTION"
.B "rpkg"
is a script to manage spec enriched Git repositories - meaning any repositories which contain
an rpm spec file. It is designed to work with expanded application sources as well as with standard DistGit
tarballs and patches. Note that you should trust the .spec files you work with because many operations
(like `rpkg srpm` or `rpkg lint`) involve parsing the spec file, which brings along evaluation
of any shell or lua scriplets present in it.
"""

man_footer = """\
.SH "CONFIGURATION"

rpkg reads configuration options from the following files in this order:

    /etc/rpkg.conf, ~/.config/rpkg.conf, <git_root>/rpkg.conf, ./rpkg.conf

Configuration found in ./rpkg.conf has the highest priority. Configuration
in /etc/rpkg.conf has the lowest priority. <git_root> placeholder above means
the top level of the Git repository we are working in. If we are in a plain
directory and all our parent directories are also plain, reading of
<git_root>/rpkg.conf is simply skipped.

.SS "DEFAULT COMMENTED CONFIGURATION"

    [rpkg]
    # whether to enable spec preprocessing,
    # that is parsing and evaluation of {{{ ... }}} expressions
    preprocess_spec = True

    # base directory where results should be stored
    base_output_path = /tmp/rpkg

    # path to user preproc macros
    # you may use ${git_root} variable
    # to anchor the path to toplevel of
    # of your git repo.
    user_macros_path =

    # optional path to rpmlint config
    # file to adjust linting of your
    # package. You may use ${git_root}
    # to anchor the path to toplevel
    # of your git repo.
    rpmlint_config =

    [git]
    # default Git repository URL, used by `rpkg clone <module>`
    clone_url = ssh://${user}@pkgs.fedoraproject.org/%(project)s

    # default anonymous Git repository URL, used by `rpkg clone -a <module>`
    anon_clone_url = https://src.fedoraproject.org/%(project)s

    [lookaside]
    # download URL for DistGit lookaside cache, used by `rpkg sources` command
    # %(ns1)s, %(name)s are parts of a module name (e.g. rpms/rpkg-util)
    # %(filename)s, %(hashtype)s, %(hash)s, %(filename)s fields are read from 'sources' file if present
    download_url = https://${git_pull_netloc}/repo/pkgs/%(ns1)s/%(name)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s

    # upload URL for DistGit lookaside cache, used by `rpkg upload`
    upload_url = https://${git_push_netloc}/repo/pkgs/upload.cgi


.SH "EXAMPLES"

Run %prep phase of an rpm build:

    $ rpkg prep
    git_pack: packing directory /home/clime/rpkg-util
    git_pack: Wrote: /tmp/rpkg/rpkg-util-1-j8x35rb3/rpkg-util-2.2.git.18.5b6a884.tar.gz
    Wrote: /tmp/rpkg/rpkg-util-1-j8x35rb3/rpkg-util.spec
    Executing(%prep): /bin/sh -e /var/tmp/rpm-tmp.vE42SH
    + umask 022
    + cd /tmp/rpkg/rpkg-util-1-j8x35rb3
    + cd /tmp/rpkg/rpkg-util-1-j8x35rb3
    + rm -rf rpkg-util-2.2.git.18.5b6a884
    + /usr/bin/gzip -dc /tmp/rpkg/rpkg-util-1-j8x35rb3/rpkg-util-2.2.git.18.5b6a884.tar.gz
    + /usr/bin/tar -xof -
    + STATUS=0
    + '[' 0 -ne 0 ']'
    + cd rpkg-util-2.2.git.18.5b6a884
    + /usr/bin/chmod -Rf a+rX,u+w,g-w,o-w .
    + exit 0

For the following command, you need to have a project in COPR created. See quick tutorial at

https://docs.pagure.org/copr.copr/user_documentation.html#quick-start

on how to do it.

    $ rpkg build
    git_pack: packing directory /home/clime/rpkg-util
    git_pack: Wrote: /tmp/rpkg/rpkg-util-2-ab6awb8y/rpkg-util-2.2.git.18.0cb6974.tar.gz
    Wrote: /tmp/rpkg/rpkg-util-2-ab6awb8y/rpkg-util.spec
    Wrote: /tmp/rpkg/rpkg-util-2-ab6awb8y/rpkg-util-2.2.git.18.0cb6974-1.fc28.src.rpm
    Uploading package /tmp/rpkg/rpkg-util-2-ab6awb8y/rpkg-util-2.2.git.18.0cb6974-1.fc28.src.rpm
    100% |################################| 122kB 359kB/s eta 0:00:00
    Build was added to rpkg-util:
      https://copr.fedorainfracloud.org/coprs/build/782327/
    Created builds: 782327

Here you can see a simple operation of generation spec file from a template:

    $ rpkg spec
    Wrote: /tmp/rpkg/rpkg-util-3-1u34plfx/rpkg-util.spec

.SH "SPEC TEMPLATES"

Apart from being useful as a means to handle unpacked as well as packed content,
they also make it possible to generate certain parts of your spec file automatically
from a Git repository where your spec file is located. You need to have "spec_preprocess"
configuration option set to True in rpkg config to use this functionality.

.SS "BASIC EXAMPLE"

    Name:       {{{ git_dir_name }}}
    Version:    {{{ git_dir_version }}}
    Release:    1%{?dist}
    Summary:    This is a test package.

    License:    GPLv2+
    URL:        https://someurl.org
    VCS:        {{{ git_dir_vcs }}}

    Source:     {{{ git_dir_pack }}}

    %description
    This is a test package.

    %prep
    {{{ git_dir_setup_macro }}}

    %changelog
    {{{ git_dir_changelog }}}

Having a spec template like this in your project (instead of just a plain spec file)
allows you to build source and rpm packages automatically following git history of your project.

.SS "MACRO REFERENCE"

.TP

BASE GIT MACROS

Those macros provide the base implementation for other macro types. They can be used directly as well. Many of them operate on "global" repository level.

Output of every git macro call is cached and other macros are made to take those cached values into account by default.
You can see that a macro uses a cached value as a default input parameter value by looking at its function header definition and searching for the `OUTPUT` keyword.

E.g.

    git_version [name="${OUTPUT[git_name]}"] [lead=0] [follow=]

Default value for name in this definition is derived from the cached output of the
previously invoked `git_name` macro. If `git_name` macro has not been invoked
or if its output was an empty string, the default value for the `name=` parameter
above will be also an empty string.

.B git_name [name=] [prepend=] [append=]

Without any parameters, lookup remote URL for the currently active
branch and output its basename with .git suffix stripped. If the
remote URL could not be determined, use "origin" remote URL
instead.

    Optional arguments:
        name     the base name to be output
        prepend  string to be prepended to the base name
        append   string to be appended to the base name

.B git_version [name="${OUTPUT[git_name]}"] [lead=0] [follow=]

Output version of a subpackage given by ${name}. The version
is a string constructed of the following parts:

    "${lead}.${follow}${commit_count_appendix}${dirty_appendix}"

${lead} is a mandatory argument of the macro. By default it is set
to zero.

${follow} (if not specified) is obtained from the follow part of
the latest tag made for the subpackage given by ${name}.

${commit_count_appendix} is generated only if the current commit
is not tagged for the given subpackage and it is composed of the
following parts

    .git.<commit_count>.<latest_commit_short_hash>

Commit count is a number of commits from the latest tag on the
given subpackage.

${dirty_appendix} is generated only if the currently checked
out working tree is dirty. It is composed of the following parts:

    .dirty.<encoded_latest_file_status_change>

    Optional arguments:
        name     name of a subpackage
        lead     major version substring
        follow   minor version substring, dynamically generated if not specified

    Environment variables:
        VERSION_BUMP: if set to a non-empty value, the following procedure is applied
        to get the version string:

        If $follow is specified by user explicitly, return $lead.$follow.

        Otherwise, extract $follow from the latest tag as the last version component (after the last dot)
        and return $lead.$follow+1. If $follow is not a number, throw an error (manual setting of follow
        is needed).

        rpkg sets the VERSION_BUMP variable into the preprocessing environment during
        execution of `rpkg tag` subcommand. Otherwise, it is unset.

.B git_vcs path=

Output pseudo repository URL pointing to the respective path in the currently checked out tree.

Example output: git+ssh://git@pagure.io/test-project.git#dadef2a8b9554e94797a7336261192e02d5d9351:subdir

    Optional arguments:
        path     path to a subtree to generate the URL for

.B git_pack path= [dir_name=] [source_name=]

Pack the whole working tree content (ommiting untracked and git-ignored content) into a gzipped source tarball and output $source_name, which is a filename of the created tarball.
Uses GNU tar to do the job. Untracked files, files ignored by git as well as .git metadata directory will be excluded from the resulting tarball.

    Optional arguments:
        path         path to a specific subdirectory to be packed
        dir_name     top-level directory name in the created tarball
        source_name  filename of the created source tarball

    Enviroment variables:
        OUTDIR: if empty, git_pack will not generate any sources, only the source filename will be output

        rpkg sets OUTDIR, only if the tarball is actually needed to be generated. It does not set it
        when `rpkg spec` is called (without any additional arguments).

.B git_archive path= [dir_name=] [source_name=]

Pack the latest commit tree content into a gzipped source tarball and output $source_name, which is a filename of the created tarball.
Uses `git archive` to do the job.

    Optional arguments:
        path         path to a specific subdirectory to be packed
        dir_name     top-level directory name in the created tarball
        source_name  filename of the created source tarball

    Enviroment variables:
        OUTDIR: if empty, git_archive will not generate any sources, only the source filename will be output

        rpkg sets OUTDIR, only if the tarball is actually needed to be generated. It does not set it
        when `rpkg spec` is called (without any additional arguments).

.B git_setup_macro [path=] [dir_name=]

Output %setup rpm macro for the given path or dir_name.

    Optional arguments:
        path         path from which dir_name is derived if not specified
        dir_name     name of the top-level directory in the source-tarball

.B git_changelog [name="${OUTPUT[git_name]}"] [since_tag=] [until_tag=] [header_locale=POSIX] [header_date_format="%a %b %d %Y"] [body_wrap=80]

Output rpm spec changelog as generated from tag messages for the subpackage given by ${name}.

    Optional arguments:
        name                name of the subpackage to generate the changelog for
        since_tag           start the changelog records with this tag
        until_tag           end the changelog records with this tag
        header_locale       locale to be used when changelog record date is being generated
        header_date_format  date format to be used in the changelog record headers
        body_wrap           maximum allowed line length for changelog body

When `rpkg tag` is called, it allows you to specify a tag message. This is the message
that will be used as a changelog record body.

.TP

GIT_REPO MACROS

These macros work on repository level and use globally available repository information. They are basically
tiny wrappers of the GIT BASE MACROS with some input values set specifically for the repository-level operation.

NOTE: If your local repository is not properly initialized or if, for example, the .git metadata directory in your local tree
has been renamed or removed, operation of the base macros can get out of the intended scope. It is therefore recommended to
use more specific git_dir_* macros where possible, which are always bound to the directory where the input spec file is
located. This is mainly relevant for git_pack and git_archive macros (see below).

git_repo_name

    git_name

git_repo_version

    git_version name="$(git_repo_name)"

git_repo_release

    git_release name="$(git_repo_name)"

git_repo_changelog

    git_changelog name="$(git_repo_name)"

git_repo_vcs

    git_vcs path="$GIT_ROOT"

git_repo_pack

    git_pack path="$GIT_ROOT"

git_repo_archive

    git_archive path="$GIT_ROOT"

git_repo_setup_macro

    git_setup_macro path="$GIT_ROOT"

.TP

GIT_DIR MACROS

These macros operate on a directory level. In particular, on a directory where the input spec file template is located. They are basically
tiny wrappers of the GIT BASE MACROS with some input values set specifically for the directory-level operation. The input spec template location is
given by INPUT_PATH environment variable set by rpkg. You can influence it from the command line by explicitly specifying --spec <template_path>
argument for rpkg subcommands that support it.

Below you can see to what the GIT DIR macros expand:

.B git_dir_name

    git_name append="$(git_prefix "$INPUT_DIR_PATH" | git_prefix_to_name_suffix)"

    Outputs repository origin URL basename appended with the path (after slash to dash substitution) from the repository
    root to the spec file's directory.

.B git_dir_version

    git_version name="$(git_dir_name)"

.B git_dir_release

    git_release name="$(git_dir_name)"

.B git_dir_changelog

    git_changelog name="$(git_dir_name)"

.B git_dir_vcs

    git_vcs path="$(dirname "$INPUT_PATH")"

    Outputs VCS pseudo URL pointing to the spec file's subdirectory.

.B git_dir_pack

    git_pack path="$(dirname "$INPUT_PATH")"

    Packs spec file's subdirectory by using GNU tar.

.B git_dir_archive

    git_archive path="$(dirname "$INPUT_PATH")"

    Packs spec file's subdirectory by using `git archive`.

.B git_dir_setup_macro

    git_setup_macro path="$(dirname "$INPUT_PATH")"

.TP

GIT_CWD MACROS

These macros operate on a directory level. In particular, on a rpkg's working directory level. Again, they
are just tiny wrappers of the GIT BASE MACROS with some input values set specifically for the working dir-level
operation.

You can influence the rpkg's working dir from the command line by explicitly specifying its --path argument.

Below you can see to what the GIT_CWD macros expand.

.B git_cwd_name

    git_name append="$(git_prefix "$PWD" | git_prefix_to_name_suffix)"

    Outputs repository origin URL basename appended with the rpkg's working directory path (after slash to dash substitution).

.B git_cwd_version

    git_version name="$(git_cwd_name)"

.B git_cwd_release

    git_release name="$(git_cwd_name)"

.B git_cwd_changelog

    git_changelog name="$(git_cwd_name)"

.B git_cwd_vcs

    git_vcs path="$PWD"

    Outputs VCS pseudo URL pointing to the rpkg's working directory.

.B git_cwd_pack

    git_pack path="$PWD"

    Packs rpkg's working directory by using GNU tar.

.B git_cwd_archive

    git_archive path="$PWD"

    Packs rpkg's working directory by using `git archive`.

.B git_cwd_setup_macro

    git_setup_macro path="$PWD"

.SH USER-DEFINED MACROS

You can define your own rpkg macros that are basically just standard bash functions
read from file specified by user_macros_path config variable. Standard output of
your custom macro will be captured and the {{{ <macro_name> }}} invocation in an
input spec file will be replaced with it.

Example of a custom macro placed in rpkg.macros file:

    function git_commits_no {
        total_commits="$(git rev-list --all --count)"
        echo -n "$total_commits"
    }

.SH "SEE ALSO"
.UR "https://pagure.io/rpkg-util/"
.BR "https://pagure.io/rpkg-util/"
"""


class ManFormatter(object):
    def __init__(self, man):
        self.man = man

    def write(self, data):
        for line in data.split('\n'):
            self.man.write('  %s\n' % line)


def strip_usage(s):
    """Strip "usage: " string from beginning of string if present"""
    if s.startswith('usage: '):
        return s.replace('usage: ', '', 1)
    else:
        return s


def man_constants():
    """Global constants for man file templates"""
    today = datetime.date.today()
    today_manstr = today.strftime(r'%Y\-%m\-%d')
    return {'today': today_manstr}


def generate(parser, subparsers):
    """\
    Generate the man page on stdout

    Given the argparse based parser and subparsers arguments, generate
    the corresponding man page and write it to stdout.
    """

    # Not nice, but works: Redirect any print statement output to
    # stderr to avoid clobbering the man page output on stdout.
    man_file = sys.stdout
    sys.stdout = sys.stderr

    mf = ManFormatter(man_file)

    choices = subparsers.choices
    k = sorted(choices.keys())

    man_file.write(man_header % man_constants())

    helptext = parser.format_help()
    helptext = strip_usage(helptext)
    helptextsplit = helptext.split('\n')
    helptextsplit = [line for line in helptextsplit
                     if not line.startswith('  -h, --help')]

    man_file.write('.SS "%s"\n' % ("Global Options",))

    outflag = False
    for line in helptextsplit:
        if line == "optional arguments:":
            outflag = True
        elif line == "":
            outflag = False
        elif outflag:
            man_file.write("%s\n" % line)

    help_texts = {}
    for pa in subparsers._choices_actions:
        help_texts[pa.dest] = getattr(pa, 'help', None)

    man_file.write('.SH "COMMAND OVERVIEW"\n')

    for command in k:
        cmdparser = choices[command]
        usage = cmdparser.format_usage()
        usage = strip_usage(usage)
        usage = ''.join(usage.split('\n'))
        usage = ' '.join(usage.split())
        if help_texts[command]:
            man_file.write('.TP\n.B "%s"\n%s\n' % (usage, help_texts[command]))
        else:
            man_file.write('.TP\n.B "%s"\n' % (usage))

    man_file.write('.SH "COMMAND REFERENCE"\n')
    for command in k:
        cmdparser = choices[command]

        man_file.write('.SS "%s"\n' % cmdparser.prog)

        help = help_texts[command]
        if help and not cmdparser.description:
            if not help.endswith('.'):
                help = "%s." % help
            cmdparser.description = help

        h = cmdparser.format_help()
        mf.write(h)

    man_file.write(man_footer)


if __name__ == '__main__':
    module_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    sys.path.insert(0, module_path)

    from rpkglib import RpkgUtil
    util = RpkgUtil()
    generate(util.parser, util.subparsers)
