import os
import sys
import getpass
import logging
import configparser
import argparse
import pprint
import git
import re
import getpass

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from munch import Munch
from cached_property import cached_property

from .utils import run
from .rpm_package import RpmPackage
from .lookaside_cache import LookasideCache
from .git_client import GitClient

log = logging.getLogger('__main__')
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler(sys.stdout))

class RpkgUtil(object):

    def main(self):
        self.setup_log()

        try:
            self.args.command()
        except Exception as e:
            if self.args.verbose:
                raise
            else:
                log.error(str(e))
        except KeyboardInterrupt:
            pass

    ######################## setup ########################

    def setup_log(self):
        if self.args.verbose:
            log.setLevel(logging.DEBUG)
        elif self.args.quiet:
            log.setLevel(logging.WARNING)
        else:
            log.setLevel(logging.INFO)

    @cached_property
    def args(self):
        return self.parser.parse_args()

    @cached_property
    def git_props(self):
        git_props = Munch()

        git_props.git_root = run(
            ['git', 'rev-parse', '--show-toplevel'],
            capture_stderr=True, throw=False)
        git_props.git_branch = run(
            ['git', 'symbolic-ref', '--short', 'HEAD'],
            capture_stderr=True, throw=False)
        git_props.git_remote = run(
            ['git', 'config', 'branch.{0}.remote'.format(git_props.git_branch)],
            capture_stderr=True, throw=False)
        git_props.git_pull_url = run(
            ['git', 'remote', 'get-url', '{0}'.format(git_props.git_remote)],
            capture_stderr=True, throw=False)
        git_props.git_push_url = run(
            ['git', 'remote', 'get-url', '--push', '{0}'.format(git_props.git_remote)],
            capture_stderr=True, throw=False)

        git_props.git_pull_netloc = urlparse(git_props.git_pull_url).netloc
        git_props.git_push_netloc = urlparse(git_props.git_push_url).netloc

        return git_props

    @cached_property
    def cfg_parser(self):
        parser = configparser.ConfigParser(
            allow_no_value=True, interpolation=configparser.ExtendedInterpolation())
        default_conf = os.path.join(os.path.dirname(__file__), 'default.conf')

        if self.args.config_file:
            parser.read([default_conf, self.args.config_file])
        else:
            glob_conf = '/etc/rpkg.conf'
            user_conf = os.path.expanduser('~/.config/rpkg.conf')
            cwd_conf = './rpkg.conf'
            if self.git_props.git_root:
                repo_conf = os.path.join(self.git_props.git_root, 'rpkg.conf')
                parser.read([default_conf, glob_conf, user_conf, repo_conf, cwd_conf])
            else:
                parser.read([default_conf, glob_conf, user_conf, cwd_conf])

        parser['DEFAULT']['user'] = self.args.user
        parser['DEFAULT']['path'] = self.args.path

        for git_prop, value in self.git_props.items():
            parser['DEFAULT'][git_prop] = value

        for key_val in self.args.config:
            key, val = re.split(r'\s*=\s*', key_val, maxsplit=1)
            if '.' in key:
                section, param = re.split(r'\.', key, maxsplit=1)
            else:
                section, param = 'DEFAULT', key
            parser[section][param] = val

        return parser

    @cached_property
    def cfg(self):
        return Munch({
            'rpkg': Munch({
                'preprocess_spec': self.cfg_parser.getboolean('rpkg', 'preprocess_spec'),
                'base_output_path': os.path.expanduser(self.cfg_parser.get('rpkg', 'base_output_path')),
                'user_macros_path': os.path.expanduser(self.cfg_parser.get('rpkg', 'user_macros_path')),
                'rpmlint_config': os.path.expanduser(self.cfg_parser.get('rpkg', 'rpmlint_config')),
                #'with': self.cfg_parser.get('rpkg', 'with'),
                #'without': self.cfg_parser.get('rpkg', 'without'),
                #'target': self.cfg_parser.get('rpkg', 'target', fallback='bar'),
                #'spec': self.cfg_parser.get('rpkg', 'spec', fallback='bar'),
                #'outdir': self.cfg_parser.get('rpkg', 'outdir', fallback='bar'),
            }),
            'git': Munch({
                'clone_url': self.cfg_parser.get('git', 'clone_url'),
                'anon_clone_url': self.cfg_parser.get('git', 'anon_clone_url'),
            }),
            'lookaside': Munch({
                'download_url': self.cfg_parser.get('lookaside', 'download_url'),
                'upload_url': self.cfg_parser.get('lookaside', 'upload_url'),
            }),
            'plugin': Munch({
                'user_plugin_path': os.path.expanduser(self.cfg_parser.get('plugin', 'user_plugin_path')),
            }),
        })

    @cached_property
    def rpm_package(self):
        return RpmPackage(self.args.path, self.args.verbose, getattr(self.args, 'spec', None),
                          getattr(self.args, 'outdir', None), **self.cfg.get('rpkg'))

    @cached_property
    def git_client(self):
        return GitClient(self.args.path, self.args.verbose, **self.cfg.get('git'))

    @cached_property
    def lookaside(self):
        return LookasideCache(self.args.path, self.args.verbose, **self.cfg.get('lookaside'))

    @cached_property
    def parser(self):
        """Setup the argument parser and register some basic commands."""
        parser = argparse.ArgumentParser(
            prog='rpkg', epilog='For detailed help pass --help to a target subcommand.')

        parser.add_argument('--config', '-c', metavar='KEY=VAL', default=[],
                            action='append', help='Set specific config '
                            'parameter KEY to a value VAL. May be specified '
                            'multiple times. Use <section>.<param> format '
                            'for KEY. Takes effect after -f option.')
        parser.add_argument('--config-file', '-f', action='store', default=None,
                            help='Specify a unique config file to be used.')
        parser.add_argument('--user', '-u', default=getpass.getuser(),
                            help='Override the discovered system user name.')
        parser.add_argument('--path', '-C', action='store', default=os.getcwd(),
                            help='Define the directory to work in (defaults to cwd).')
        parser.add_argument('--verbose', '-v', action='store_true',
                            help='Run with verbose debug output.')
        parser.add_argument('--quiet', '-q', action='store_true',
                            help='Run quietly only displaying errors.')

        #parser.set_defaults(command=parser.print_help)
        self.subparsers = self.setup_subparsers(parser)

        return parser

    def setup_subparsers(self, parser):
        sp = parser.add_subparsers(
            help='Available subcommands')

        self.register_copr_build(sp)
        self.register_changes(sp)
        self.register_clean(sp)
        self.register_clog(sp)
        self.register_clone(sp)
        self.register_config(sp)
        self.register_commit(sp)
        self.register_compile(sp)
        self.register_diff(sp)
        self.register_gimmespec(sp)
        self.register_import_srpm(sp)
        self.register_install(sp)
        self.register_lint(sp)
        self.register_local(sp)
        self.register_prep(sp)
        self.register_pull(sp)
        self.register_push(sp)
        self.register_sources(sp)
        self.register_spec(sp)
        self.register_srpm(sp)
        self.register_switch_branch(sp)
        self.register_tag(sp)
        self.register_unused_patches(sp)
        self.register_upload(sp)
        self.register_verify_files(sp)
        self.register_nvr(sp)

        return sp

    ######################## commands ########################

    def copr_build(self):
        print("N/A")
        return
        srpm_path = self.rpm_package.srpm()

        if self.args.project:
            project = self.args.project
        else:
            project = self.git_client.push_url_path

        self.copr_build(project,
                        srpm_path,
                        not self.args.wait,
                        self.args.copr_config)

    def config(self):
        self.cfg_parser.write(sys.stdout)

    def changes(self):
        print("N/A")
        return
        pass

    def clean(self):
        self.git_client.clean(self.args.dry_run, self.args.x)

    def clog(self):
        print("N/A")
        return
        self.rpm_package.clog(self.args.raw)

    def clone(self):
        print("N/A")
        return
        pass

    def commit(self):
        print("N/A")
        return
        if self.args.with_changelog and not self.args.message:
            raise RpkgError('--with-changelog must be used with -m together.')

        if self.args.message and self.args.with_changelog:
            # Compose commit message with a summary and content into a file.
            self.rpm_package.clog(raw=True)
            clog_file = os.path.abspath(os.path.join(self.args.path, 'clog'))
            commit_msg_file = os.path.abspath(os.path.join(self.args.path, 'commit-message'))
            with open(commit_msg_file, 'w') as commit_msg:
                commit_msg.write(self.args.message)
                commit_msg.write('\n\n')
                with open(clog_file, 'r') as clog:
                    commit_msg.write(clog.read())
            self.args.file = commit_msg_file
            os.remove(clog_file)
            # This assignment is a magic because commit message is in the file
            # commit-message already.
            self.args.message = None

        try:
            self.git_client.commit(
                self.args.message, self.args.file, self.args.files, self.args.signoff)

            if self.args.tag:
                self.rpm_package.dump_spec_only = True
                self.rpm_package.version_bump = True
                self.rpm_package.release_bump = True
                self.git_client.add_tag()

        except Exception:
            if self.args.tag:
                self.log.error('Could not commit, will not tag!')
            if self.args.push:
                self.log.error('Could not commit, will not push!')
            raise

        finally:
            if self.args.with_changelog and os.path.isfile(self.args.file):
                os.remove(self.args.file)
                del self.args.file

        if self.args.push:
            self.push()

    def compile(self):
        #self.lookaside.sources()
        self.rpm_package.compile(target=self.args.target,
                         short=self.args.short_circuit,
                         nocheck=self.args.nocheck,
                         bcond_with=getattr(self.args, 'with'),
                         bcond_without=self.args.without)

    def diff(self):
        print("N/A")
        return
        pass

    def gimmespec(self):
        print(self.rpm_package.repo_spec)

    def import_srpm(self):
        print("N/A")
        return
        pass

    def install(self):
        #self.lookaside.sources()
        self.rpm_package.install(target=self.args.target,
                                 short=self.args.short_circuit,
                                 nocheck=self.args.nocheck,
                                 bcond_with=getattr(self.args, 'with'),
                                 bcond_without=self.args.without)

    def lint(self):
        self.rpm_package.use_existing_outdir = True
        self.rpm_package.lint(self.args.info, self.args.rpmlint_config)

    def local(self):
        #self.lookaside.sources()
        self.rpm_package.local(target=self.args.target,
                               bcond_with=getattr(self.args, 'with'),
                               bcond_without=self.args.without)

    def prep(self):
        #self.lookaside.sources()
        self.rpm_package.prep(target=self.args.target,
                      bcond_with=getattr(self.args, 'with'),
                      bcond_without=self.args.without)

    def pull(self):
        print("N/A")
        return
        self.git_client.pull(self.args.rebase, self.args.no_rebase)

    def push(self):
        print("N/A")
        return
        self.git_client.push(self.args.force)

    def sources(self):
        print("N/A")
        return
        pass

    def spec(self):
        self.rpm_package.dump_spec_only = not self.args.sources
        self.rpm_package.load_spec()

    def srpm(self):
        #self.lookaside.sources()
        return self.rpm_package.srpm(target=getattr(self.args, 'target', None),
                                     bcond_with=getattr(self.args, 'with', []),
                                     bcond_without=getattr(self.args, 'without', []))

    def switch_branch(self):
        print("N/A")
        return
        pass

    def tag(self):
        if self.args.list:
            print("N/A")
            return
            self.git_client.list_tag(self.args.tag)
            return
        elif self.args.delete:
            print("N/A")
            return
            if not self.args.tag:
                sys.stderr.write('Please, specify tag to be deleted.\n')
            else:
                self.git_client.delete_tag(self.args.tag)
            return

        tag = self.args.tag
        self.rpm_package.dump_spec_only = True

        if not tag:
            self.rpm_package.version_bump = True
            self.rpm_package.release_bump = True
            self.rpm_package._rpmdefines = self.rpm_package.rpmdefines + ['--define', 'dist %nil']
            tag = self.rpm_package.nvr

        self.git_client.add_tag(tag, self.args.skip_edit)

        self.rpm_package.version_bump = False
        self.rpm_package.release_bump = False
        self.rpm_package.load_spec()

    def unused_patches(self):
        print("N/A")
        return
        self.rpm_package.dump_spec_only = True
        unused = self.rpm_package.unused_patches()
        print('\n'.join(unused))

    def upload(self):
        print("N/A")
        return
        pass

    def verify_files(self):
        self.rpm_package.use_existing_outdir = True
        self.rpm_package.verify_files()

    def nvr(self):
        self.rpm_package.dump_spec_only = True
        print('%s-%s-%s' % (self.rpm_package.module_name, self.rpm_package.ver,
                            self.rpm_package.rel))

    ######################## subparsers ########################

    def register_copr_build(self, subparsers):
        """Register the copr build target"""
        copr_build_parser = subparsers.add_parser(
            'copr-build', parents=[self.spec_common_parser],
            description='Build package in Copr Build System. '
            'You need to have copr-cli installed and also set up '
            'Copr api key. For more information about the api key, '
            'see API KEY section in copr-cli(1) man page.',
            help='N/A Build package in Copr Build System')
        copr_build_parser.add_argument(
            '--config', required=False,
            metavar='CONFIG', dest='build_config',
            help="Path to an alternative Copr configuration file")
        copr_build_parser.add_argument(
            '--wait', '-w', action='store_true', default=False,
            help='Wait for build to finish.')
        copr_build_parser.add_argument(
            'project', nargs='?',
            help='Name of the project in format OWNER/PROJECT. If not '
            'specified, project name will be determined from git push '
            'url path.')
        copr_build_parser.set_defaults(command=self.copr_build)

    def register_config(self, subparsers):
        config_parser = subparsers.add_parser(
            'config', description='Do nothing, only output used config.',
            help='Do nothing, only output used config.')
        config_parser.set_defaults(command=self.config)

    def register_changes(self, subparsers):
        """Register the changes target"""
        changes_parser = subparsers.add_parser(
            'changes', parents=[self.spec_common_parser],
            description='This will use git to show a diff '
            'of all the changes (even uncommitted changes) '
            'since the last package tag was created.',
            help='N/A Diff against last package tag')
        changes_parser.set_defaults(command=self.changes)

    def register_clean(self, subparsers):
        """Register the clean target"""
        clean_parser = subparsers.add_parser(
            'clean', description='This command can be used to '
            'clean up your working directory. By default it will '
            'follow .gitignore rules.',
            help='Remove untracked files.')
        clean_parser.add_argument(
            '--dry-run', '-n', action='store_true',
            help='Don\'t actually remove anything, just show '
            'what would be done.')
        clean_parser.add_argument(
            '-x', action='store_true',
            help='Do not follow .gitignore rules')
        clean_parser.set_defaults(command=self.clean)

    def register_clog(self, subparsers):
        """Register the clog target"""
        clog_parser = subparsers.add_parser(
            'clog', description='This will create a file named "clog" '
            'that contains the latest rpm changelog entry. The leading '
            '"- " text will be stripped.',
            help='N/A Make a clog file containing top changelog entry.')
        clog_parser.add_argument(
            '--raw', action='store_true',
            help='Generate a more "raw" clog without '
            'twiddling the contents')
        clog_parser.set_defaults(command=self.clog)

    def register_clone(self, subparsers):
        """Register the clone target"""
        clone_parser = subparsers.add_parser(
            'clone', description='This command will clone the named '
            'repository from the configured repository base URL. '
            'By default it will also checkout the master branch '
            'for your working copy.',
            help='N/A Clone and checkout a repository.')
        clone_parser.add_argument(
            'repo', action='store',
            help='Name of the repository to clone')
        clone_parser.add_argument(
            'clone_target', action='store',
            help='Directory in which to clone the repository')
        clone_parser.add_argument(
            '--branches', '-B', action='store_true',
            help='Do an old style checkout with subdirs for branches')
        clone_parser.add_argument(
            '--branch', '-b', action='store',
            help='Check out a specific branch')
        clone_parser.add_argument(
            '--anonymous', '-a', action='store_true',
            help='Check out a repository anonymously')
        clone_parser.set_defaults(command=self.clone)

    def register_commit(self, subparsers):
        """Register the commit target and ci alias"""
        commit_parser = subparsers.add_parser(
            'commit', description='This invokes a git commit. '
            'All tracked files with changes will be committed '
            'unless a specific file list is provided. $EDITOR '
            'will be used to generate a changelog message '
            'unless one is given to the command. A tag or '
            'push can be done at the same time.',
            help='N/A Commit changes')
        commit_parser.add_argument(
            '-m', '--message', default=None,
            help='Use the given <msg> as the commit message summary')
        commit_parser.add_argument(
            '--with-changelog', action='store_true', default=None,
            help='Get the last changelog from SPEC as commit message '
            'content. This option must be used with -m together.')
        commit_parser.add_argument(
            '-t', '--tag', default=False, action='store_true',
            help='Create a tag for this commit')
        commit_parser.add_argument(
            '-F', '--file', default=None,
            help='Take the commit message from the given file')
        commit_parser.add_argument(
            '-p', '--push', default=False, action='store_true',
            help='Commit and push as one action')
        commit_parser.add_argument(
            'files', nargs='*', default=[],
            help='Optional list of specific files to commit')
        commit_parser.add_argument(
            '-s', '--signoff', default=False, action='store_true',
            help='Include a signed-off-by')
        commit_parser.set_defaults(command=self.commit)

        ci_parser = subparsers.add_parser(
            'ci', parents=[commit_parser], conflict_handler='resolve',
            help='Alias for commit')
        ci_parser.set_defaults(command=self.commit)

    def register_compile(self, subparsers):
        """Register the compile target"""
        compile_parser = subparsers.add_parser(
            'compile', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='This command calls rpmbuild to compile the source. '
            'By default the prep and configure stages will be done as well, '
            'unless the short-circuit option is used.',
            help='Local test rpmbuild compile.')
        compile_parser.add_argument(
            '--short-circuit', action='store',
            help='short-circuit compile')
        compile_parser.add_argument(
            '--nocheck', action='store',
            help='nocheck compile')
        compile_parser.set_defaults(command=self.compile)

    def register_diff(self, subparsers):
        """Register the diff target"""
        diff_parser = subparsers.add_parser(
            'diff', description='Use git diff to show changes that have been '
            'made to tracked files. By default cached changes (changes that '
            'have been git added) will not be shown.',
            help='N/A Show changes between commits, commit and working tree, etc.')
        diff_parser.add_argument(
            'files', action='store', nargs='*', default=[],
            help='Optionally diff specific files')
        diff_parser.add_argument(
            '--cached', action='store_true',
            help='View staged changes')
        diff_parser.set_defaults(command=self.diff)

    def register_gimmespec(self, subparsers):
        """Register the gimmespec target"""
        gimmespec_parser = subparsers.add_parser(
            'gimmespec', description='Print auto-located input spec file name',
            help='Print auto-located input spec file name')
        gimmespec_parser.set_defaults(command=self.gimmespec)

    def register_import_srpm(self, subparsers):
        """Register the import target"""
        import_parser = subparsers.add_parser(
            'import', description='This will extract sources, patches, '
            'and the spec file from an srpm and update the current module '
            'accordingly. It will import to the current branch by default.',
            help='N/A Import srpm content into a module')
        import_parser.add_argument(
            'srpm', action='store',
            help='Source rpm to import')
        import_parser.add_argument(
            '--skip-diffs', action='store_true',
            help="Don't show diffs when importing srpms")
        import_parser.set_defaults(command=self.import_srpm)

    def register_install(self, subparsers):
        """Register the install target"""
        install_parser = subparsers.add_parser(
            'install', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='This will call rpmbuild to run the install section. '
            'All leading sections will be processed as well, unless the '
            'short-circuit option is used.',
            help='Local test rpmbuild install.')
        install_parser.add_argument(
            '--short-circuit', action='store',
            help='short-circuit install')
        install_parser.add_argument(
            '--nocheck', action='store',
            help='nocheck install')
        install_parser.set_defaults(command=self.install)

    def register_lint(self, subparsers):
        """Register the lint target"""
        lint_parser = subparsers.add_parser(
            'lint', description='Rpmlint can be configured using '
            'the --rpmlint-config/-r option or by setting rpkg.rpmlint_config'
            'variable.',
            help='Run rpmlint against local spec '
            'and build output if present.')
        lint_parser.add_argument(
            '--info', '-i', default=False, action='store_true',
            help='Display explanations for reported messages')
        lint_parser.add_argument(
            '--rpmlint-config', '-r', default=None,
            help='Use a specific configuration file for rpmlint')
        lint_parser.add_argument(
            '--outdir', default=None,
            help='Directory with rpm items to be linted or just '
            'a single rpm item like a .spec file. If not specified, '
            'a latest package directory from base output path will '
            'be selected.')
        lint_parser.set_defaults(command=self.lint)

    def register_local(self, subparsers):
        """Register the local target"""
        local_parser = subparsers.add_parser(
            'local', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Locally test run of rpmbuild producing binary RPMs. '
            'The rpmbuild output will be logged into a file named '
            '.build-%{version}-%{release}.log',
            help='Local test rpmbuild binary')
        local_parser.set_defaults(command=self.local)

    def register_prep(self, subparsers):
        """Register the prep target"""
        prep_parser = subparsers.add_parser(
            'prep', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Use rpmbuild to "prep" the sources (unpack the source '
            'archive(s) and apply any patches).',
            help='prep test rpmbuild binary')
        prep_parser.set_defaults(command=self.prep)

    def register_pull(self, subparsers):
        """Register the pull target"""
        pull_parser = subparsers.add_parser(
            'pull', description='This command uses git to fetch remote '
            'changes and apply them to the current working copy. A rebase '
            'option is available which can be used to avoid merges.',
            help='N/A Pull changes from the remote repository '
            'and update the working copy.')
        pull_parser.add_argument(
            '--rebase', action='store_true',
            help='Rebase the locally committed changes on top of the remote '
            'changes after fetching. This can avoid a merge commit, but does '
            'rewrite local history.')
        pull_parser.add_argument(
            '--no-rebase', action='store_true',
            help='Do not rebase, overriding .git settings to the contrary')
        pull_parser.set_defaults(command=self.pull)

    def register_push(self, subparsers):
        """Register the push target"""
        push_parser = subparsers.add_parser(
            'push', description='Push changes and annotated tags '
            'to remote repository.',
            help='N/A Push changes and annotated tags to remote repository.')
        push_parser.add_argument(
            '--force', '-f', action='store_true',
            help='Force push')
        push_parser.set_defaults(command=self.push)

    def register_sources(self, subparsers):
        """Register the sources target"""
        sources_parser = subparsers.add_parser(
            'sources', description='Download source files from '
            'lookaside cache.',
            help='N/A Download source files from lookaside cache.')
        sources_parser.add_argument(
            '--outdir', action='store',
            help='Directory to download files into (defaults to cwd)')
        sources_parser.set_defaults(command=self.sources)

    def register_spec(self, subparsers):
        """Register the spec target"""
        spec_parser = subparsers.add_parser(
            'spec', parents=[self.spec_common_parser],
            description='Preprocess a given spec template '
            'and put the resulting spec file into outdir. The spec '
            'template is simply copied into the outdir if '
            'spec preprocessing is turned off or if the input spec '
            'template does not contain any preprocessor macros.',
            help='Generate spec file from a spec template')
        spec_parser.add_argument(
            '--sources', action='store_true', default=False,
            help='Enable source generation by the input '
            'spec template. By default disabled.')
        spec_parser.set_defaults(command=self.spec)

    def register_srpm(self, subparsers):
        """Register the srpm target"""
        srpm_parser = subparsers.add_parser(
            'srpm', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Create a source rpm',
            help='Create a source rpm')
        srpm_parser.set_defaults(command=self.srpm)

    def register_switch_branch(self, subparsers):
        """Register the switch-branch target"""
        switch_branch_parser = subparsers.add_parser(
            'switch-branch', description='This command can switch '
            'to a local git branch. If provided with a remote branch name '
            'that does not have a local match it will create one. It can '
            'also be used to list the existing local and remote branches.',
            help='N/A Work with branches')
        switch_branch_parser.add_argument(
            'branch', action='store',
            help='Branch name to switch to')
        switch_branch_parser.add_argument(
            '--list', '-l', action='store',
            help='List both remote-tracking branches and local branches')
        switch_branch_parser.add_argument(
            '--fetch', action='store',
            help='Fetch new data from remote before switch')
        switch_branch_parser.set_defaults(command=self.switch_branch)

    def register_tag(self, subparsers):
        """Register the tag target"""
        tag_parser = subparsers.add_parser(
            'tag', description='This command uses git to '
            'create, list, or delete tags.',
            help='N/A Management of git tags')
        tag_parser.add_argument(
            '-l', '--list', default=False, action='store_true',
            help='List all tags with a given pattern, or '
            'all if not pattern is given')
        tag_parser.add_argument(
            '-d', '--delete', default=False, action='store_true',
            help='Delete a tag')
        tag_parser.add_argument(
            '-s', '--skip-edit', default=False, action='store_true',
            help='Skips changelog editing when creating a tag.')
        tag_parser.add_argument(
            '-n', '--name', default='', action='store',
            help='Set package name part of the tag name to specific value.')
        tag_parser.add_argument(
            '-v', '--version', default='', action='store',
            help='Set version part of the tag name to specific value.')
        tag_parser.add_argument(
            '-r', '--release', default='', action='store',
            help='Set release part of the tag name to specific value.')
        tag_parser.add_argument(
            'tag', nargs='?', default=None, help='Name of the tag')
        tag_parser.set_defaults(command=self.tag)

    def register_unused_patches(self, subparsers):
        """Register the unused-patches target"""
        unused_patches_parser = subparsers.add_parser(
            'unused-patches', description='Print list of patches not referenced '
            'by name in the specfile.',
            help='N/A Print list of patches not referenced by name in the specfile')
        unused_patches_parser.set_defaults(command=self.unused_patches)

    def register_upload(self, subparsers):
        """Register the upload target"""
        upload_parser = subparsers.add_parser(
            'upload', description='This command will upload new source '
            'file(s) to lookaside cache. Source file names are appended '
            'to sources file, and .gitignore will be also updated with '
            'new uploaded file(s). Please remember to commit them.',
            help='N/A Upload source files')
        upload_parser.add_argument(
            'files', nargs='*', default=[],
            help='List of files to upload')
        upload_parser.set_defaults(command=self.upload)

    def register_verify_files(self, subparsers):
        """Register the verify-files target"""
        verify_files_parser = subparsers.add_parser(
            'verify-files', description='Locally run \'rpmbuild -bl\' to '
            'verify the spec file\'s %%files sections. This requires a '
            'successful run of the install commaned in advance.',
            help='Locally verify %%files section')
        verify_files_parser.add_argument(
            '--outdir', default=None,
            help='Output directory with build results to verify.')
        verify_files_parser.set_defaults(command=self.verify_files)

    def register_nvr(self, subparsers):
        """Register the nvr target"""
        nvr_parser = subparsers.add_parser(
            'nvr', parents=[self.spec_common_parser],
            description='Print package name-version-release',
            help='Print package name-version-release')
        nvr_parser.set_defaults(command=self.nvr)

    @cached_property
    def rpm_common_parser(self):
        """Create a common parser for rpm commands"""
        rpm_common_parser = argparse.ArgumentParser(
            'rpm_common', add_help=False)
        rpm_common_parser.add_argument(
            '--target', default=None,
            help='Prep for a specific arch or arch-vendor-os.')
        rpm_common_parser.add_argument(
            '--with', metavar='OPTION', action='append', default=[],
            help='Enable %%bcond spec OPTION. '
            'Can be specified multiple times.')
        rpm_common_parser.add_argument(
            '--without', metavar='OPTION', action='append', default=[],
            help='Disable %%bcond spec OPTION. '
            'Can be specified multiple times.')
        return rpm_common_parser

    @cached_property
    def spec_common_parser(self):
        """Create a common parser for commands that need a spec file"""
        spec_common_parser = argparse.ArgumentParser(
            'spec_common', add_help=False)
        spec_common_parser.add_argument(
            '--outdir', action='store', default=None,
            help='Where to put any generated rpm content. Directory must exist.')
        spec_common_parser.add_argument(
            '--spec', action='store', default=None,
            help='An input spec template. If not given, the spec template is '
            'auto-located in path as a file ending with ".spec.rpkg" '
            'or ".spec" suffix in this order.')
        return spec_common_parser
