import git
import logging
import re

from cached_property import cached_property

from . import utils

from .utils import run
from .errors import RpkgError

log = logging.getLogger('__main__')

class GitClient(object):

    def __init__(self, path, verbose, clone_url=None, anon_clone_url=None):
        self.path = path
        self.clone_url = clone_url
        self.anon_clone_url = anon_clone_url
        self.verbose = verbose
        self.quiet = False

    @cached_property
    def repo(self):
        """Create a repo object from our path"""
        log.debug('Creating repo object from %s', self.path)
        try:
            try:
                return git.Repo(self.path, search_parent_directories=True)
            except TypeError:
                # fix for old GitPython, which hardcoded True
                # for the search_parent_directories
                return git.Repo(self.path)
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            raise RpkgError('%s is not a valid repo' % self.path)

    def clean(self, dry_run, x):
        options = ['-f', '-d']

        if dry_run:
            options.append('-n')
        if x:
            options.append('-x')
        if self.quiet:
            options.append('-q')

        self.repo.git.clean(options)

    def add_tag(self, tagname, skip_edit=False):
        """Add a git tag to the repository

        Takes an optional tagname.
        Runs the tag command and returns nothing.
        """
        if not re.match(r'^.*-[^-]+-[^-]+$', tagname):
            raise RpkgError("Tag {0} does not have n-v-r format.".format(tagname))

        if tagname in self.repo.tags:
            raise RpkgError("Tag {0} already exists.".format(tagname))

        cmd = ['git', 'tag', '-a', '-m',
               self.get_tag_message(tagname, skip_edit),
               tagname]

        run(cmd, cwd=self.path)
        log.info('Created tag: %s', tagname)

    @property
    def ns_module_name(self):
        if not self._ns_module_name:
            self.load_ns_module_name()
        return self._ns_module_name

    @ns_module_name.setter
    def ns_module_name(self, ns_module_name):
        self._ns_module_name = ns_module_name

    def load_ns_module_name(self):
        """
        Loads the namespaced module name.
        """
        try:
            push_url = self.push_url
        except RpkgError:
            raise RpkgError('Unable to find push URL. Use --module-name.')

        replacements = {'user': self.user, 'module': '(.*)/?'}
        gitbaseurl_pattern = self.gitbaseurl%replacements + '$'
        anongiturl_pattern = self.anongiturl%replacements + '$'

        match = re.match(gitbaseurl_pattern, push_url)
        if not match:
            match = re.match(anongiturl_pattern, push_url)

        if match:
            ns_module_name = match.group(1)
        else:
            ns_module_name = urlparse(push_url).path.rsplit(':')[-1].strip('/')

        if ns_module_name.endswith('.git'):
            ns_module_name = ns_module_name[:-len('.git')]

        self._ns_module_name = ns_module_name

    def get_latest_merged_tagname(self, new_tagname):
        module_name = new_tagname.rsplit('-', 2)[0]
        raw_tagnames = self.repo.git.for_each_ref('--sort=-taggerdate').split('\n')
        tagnames = [re.sub(r'.*/([^/]*)', r'\1', raw_tagname) for raw_tagname in raw_tagnames]

        for tagname in tagnames:
            expr = r'^{0}-[^-]+-[^-]+$'.format(module_name)
            if not re.match(expr, tagname):
                continue

            merge_base = self.repo.git.merge_base(tagname, 'HEAD')
            tag_rev = self.repo.git.rev_parse(tagname+'^{commit}')

            if merge_base == tag_rev:
                return tagname

        return None

    def get_tag_message(self, new_tagname, skip_edit=False):
        latest_merged_tagname = self.get_latest_merged_tagname(new_tagname)

        if latest_merged_tagname:
            rev_list_expr = '{0}..HEAD'.format(
                latest_merged_tagname)
        else:
            rev_list_expr = 'HEAD'

        entries = [utils.get_changelog_entry(commit.message) for commit in
                   self.repo.iter_commits(rev_list_expr, paths=[self.path])]

        editor = self.repo.git.var('GIT_EDITOR')
        subject = 'Commits: {0}'.format(rev_list_expr)

        help_text = ("#\n"
        "# Write a changelog for tag:\n"
        "#   " + new_tagname + "\n"
        "# Lines starting with '#' will be ignored.\n")

        body = '\n'.join(entries) + '\n' + help_text
        if not skip_edit:
            body = utils.edit(editor, body)

        return '{0}\n\n{1}'.format(subject, body)

    def push(self, force=False):
        """Push changes to the remote repository"""
        self.check_repo(is_dirty=False, all_pushed=False)

        # see if our branch is tracking anything
        try:
            self.load_branch_merge()
        except Exception:
            log.warning('Current branch cannot be pushed anywhere!')

        untracked_patches = self.find_untracked_patches()
        if untracked_patches:
            log.warning(
                'Patches %s %s not tracked within either git or sources',
                ', '.join(untracked_patches),
                'is' if len(untracked_patches) == 1 else 'are')

        cmd = ['git', 'push', '--follow-tags']
        if self.quiet:
            cmd.append('-q')
        run(cmd, cwd=self.path)

    def check_repo(self, is_dirty=True, all_pushed=True):
        """Check various status of current repository

        :param bool is_dirty: Default to True. To check whether there is uncommitted changes.
        :param bool all_pushed: Default to True. To check whether all changes are pushed.
        :raises RpkgError: if any unexpected status is detected. For example,
            if changes are not committed yet.
        """
        if is_dirty:
            if self.repo.is_dirty():
                raise RpkgError('%s has uncommitted changes.  Use git status '
                                'to see details' % self.path)
        if all_pushed:
            branch = self.repo.active_branch
            try:
                remote = self.repo.git.config('--get', 'branch.%s.remote' % branch)
                merge = self.repo.git.config('--get', 'branch.%s.merge' % branch).replace(
                    'refs/heads', remote)
            except git.GitCommandError:
                raise RpkgError('Branch {0} does not track remote branch.\n'
                                'Use the following command to fix that:\n'
                                '    git branch -u origin/REMOTE_BRANCH_NAME'.format(branch))
            if self.repo.git.rev_list('%s...%s' % (merge, branch)):
                raise RpkgError('There are unpushed changes in your repo')
