import os
import logging
import subprocess
import tempfile
import glob
import rpm

from pipes import quote

from .utils import run
from .errors import RpkgError, RpmSpecParseError

log = logging.getLogger('__main__')

class RpmPackage(object):

    def __init__(self, path, verbose=False, spec=None, outdir=None,
                 preprocess_spec=True, base_output_path='/tmp/rpkg',
                 user_macros_path=None, rpmlint_config=None):

        self.path = path
        self.preprocess_spec = preprocess_spec
        self.base_output_path = base_output_path
        self.user_macros_path = user_macros_path
        self.rpmlint_config = rpmlint_config
        self.verbose = verbose
        self.quiet = False

        self.use_existing_outdir = False
        self.dump_spec_only = False
        self.version_bump = False
        self.release_bump = False
        self.use_existing_outdir = False

        self._outdir = outdir
        self._rpmdefines = []

        if spec:
            self._specloc = spec
        else:
            self._specloc = self.path

        self._repo_spec = None
        self._spec = None

        self._module_name_spec = None
        self._epoch = None
        self._ver = None
        self._rel = None
        self._nvr = None

        if not os.path.isdir(self.base_output_path):
            prev_umask = os.umask(0)
            os.makedirs(self.base_output_path, mode=0o1777)
            os.umask(prev_umask)

    @property
    def outdir(self):
        if not self._outdir:
            self.load_outdir()
        return self._outdir

    @property
    def rpmdefines(self):
        if not self._rpmdefines:
            self.load_rpmdefines()
        return self._rpmdefines

    @property
    def repo_spec(self):
        if not self._repo_spec:
            self.load_repo_spec()
        return self._repo_spec

    @property
    def spec(self):
        if not self._spec:
            self.load_spec()
        return self._spec

    @property
    def module_name(self):
        if not self._module_name_spec:
            self.load_nameverrel()
        return self._module_name_spec

    @property
    def ver(self):
        if not self._ver:
            self.load_nameverrel()
        return self._ver

    @property
    def rel(self):
        if not self._rel:
            self.load_nameverrel()
        return self._rel

    @property
    def nvr(self):
        if not self._nvr:
            self.load_nvr()
        return self._nvr

    def load_nvr(self):
        self._nvr = '%s-%s-%s' % (self.module_name, self.ver, self.rel)

    def load_repo_spec(self):
        if not os.path.isdir(self._specloc):
            self._repo_spec = os.path.abspath(self._specloc)
            return

        files = os.listdir(self._specloc)
        spec = None

        for f in files:
            if f.endswith('.spec') and not f.startswith('.'):
                spec = os.path.abspath(os.path.join(self._specloc, f))

        if self.preprocess_spec:
            for f in files:
                if f.endswith('.spec.rpkg') and not f.startswith('.'):
                    spec = os.path.abspath(os.path.join(self._specloc, f))

        if not spec:
            raise RpkgError('No spec file found at {0}.'
                            .format(self._specloc))

        self._repo_spec = spec

    def load_spec(self):
        self._spec = self.preprocess(self.repo_spec)
        if not self.dump_spec_only:
            self.setup_source_symlinks()

    def load_outdir(self):
        dirname_base = os.path.splitext(
            os.path.splitext(os.path.basename(self.repo_spec))[0])[0]
        cnt = len(glob.glob(os.path.join(self.base_output_path, dirname_base+'*')))

        if not self.use_existing_outdir:
            prefix = dirname_base + '-' + str(cnt+1) + '-'
            self._outdir = tempfile.mkdtemp(dir=self.base_output_path, prefix=prefix)
            return

        try:
            prefix = dirname_base + '-' + str(cnt) + '-'
            self._outdir = glob.glob(os.path.join(self.base_output_path, prefix+'*'))[0]
        except IndexError:
            raise RpkgError('Could not load the latest outdir with prefix "%s" at %s'
                            % (prefix, self.base_output_path))

    def set_specloc(self, specloc):
        if not specloc:
            self._specloc = self.path
        else:
            self._specloc = specloc

    def load_rpmdefines(self):
        self._rpmdefines = [
            '--define', '_sourcedir %s' % self.outdir,
            '--define', '_specdir %s' % self.outdir,
            '--define', '_builddir %s' % self.outdir,
            '--define', '_buildrootdir %s' % self.outdir,
            '--define', '_srcrpmdir %s' % self.outdir,
            '--define', '_rpmdir %s' % self.outdir
        ]

    def install(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpm -bi on a module

        optionally for a specific target, short-circuit it, or
        specify with/without options

        Logs the output and returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])

        if short:
            cmd.extend(['--short-circuit'])

        if nocheck:
            cmd.extend(['--nocheck'])

        if self.quiet:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bi', os.path.join(self.path, self.spec)])
        run(cmd, capture=False)

    def compile(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpmbuild -bc on a module

        Optionally for a specific target, short-circuit it, or
        specify with/without options.

        Logs the output and returns nothing.
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])

        if short:
            cmd.extend(['--short-circuit'])

        if nocheck:
            cmd.extend(['--nocheck'])

        if self.quiet:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bc', os.path.join(self.path, self.spec)])
        run(cmd, capture=False)

    def prep(self, target=None, bcond_with=[], bcond_without=[]):
        """Run rpm -bp on a module

        Optionally for a specific target, short-circuit it, or
        specify with/without options.

        Logs the output and returns nothing.
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])

        if self.quiet:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bp', os.path.join(self.path, self.spec)])
        run(cmd, capture=False)

    def lint(self, info=False, rpmlint_config=None):
        """Run rpmlint over built files

        Log the output and returns nothing

        rpmlint_config is the name of the config file passed to
        rpmlint if specified by the command line argument.
        """
        cmd = ['rpmlint']

        if info:
            cmd.extend(['-i'])

        if rpmlint_config:
            cmd.extend(["-f", os.path.join(self.path, rpmlint_config)])
        elif self.rpmlint_config:
            cmd.extend(["-f", os.path.join(self.path, self.rpmlint_config)])

        cmd.extend([self.outdir])
        run(cmd, capture=False)

    def local(self, target=None, bcond_with=[], bcond_without=[]):
        """rpmbuild locally for given target.

        Writes output to a log file and logs it to the logger
        Returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])

        if self.quiet:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-ba', os.path.join(self.path, self.spec)])
        run(cmd, capture=False)

    def srpm(self, target=None, bcond_with=[], bcond_without=[]):
        """Create an srpm using hashtype from content in the module

        Requires sources already downloaded.
        """
        srpm_path = os.path.join(self.outdir, "%s-%s-%s.src.rpm" %
                                 (self.module_name, self.ver, self.rel))

        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])

        if self.quiet:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bs', self.spec])
        run(cmd, capture=False)

        return srpm_path

    def verify_files(self):
        """
        Run rpmbuild -bl on a module to verify the %files section.
        """
        log.info('Working on {0}...'.format(self.outdir))

        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if self.quiet:
            cmd.extend(['--quiet'])

        try:
            spec_path = glob.glob(os.path.join(self.outdir, '*.spec'))[0]
        except IndexError:
            raise RpkgError('No spec file found at {0}'.format(self.outdir))

        cmd.extend(['-bl', spec_path])
        run(cmd, capture=False)

    def setup_source_symlinks(self):
        ts = rpm.ts()

        try:
            rpm_spec = ts.parseSpec(self.spec)
            rpm.reloadConfig()
        except ValueError as e:
            raise RpmSpecParseError(str(e))

        for (filepath, num, flags) in rpm_spec.sources:
            filename = os.path.basename(filepath)
            symlink_source = os.path.join(self.path, filename)
            symlink_dest = os.path.join(self.outdir, filename)

            if os.path.isfile(symlink_source) and not \
                    os.path.isfile(symlink_dest):
                os.symlink(symlink_source, symlink_dest)

    def get_preproc_source_params(self):
        source_params = '-s /usr/lib/rpkg.macros.d/all.bash'

        if self.user_macros_path:
            if not os.path.exists(self.user_macros_path):
                raise RpkgError('{0} does not exist.'.format(self.user_macros_path))
            source_params += ' -s {0}'.format(quote(self.user_macros_path))

        return source_params

    def get_preproc_env_params(self, spec_templ):
        spec_templ_dir = os.path.dirname(spec_templ)

        env_params = '-e INPUT_PATH={0} -e INPUT_DIR_PATH={1}'.format(
            quote(spec_templ), quote(spec_templ_dir))

        if not self.dump_spec_only:
            env_params += ' -e OUTDIR={0}'.format(quote(self.outdir))

        if self.version_bump:
            env_params += ' -e VERSION_BUMP=1'

        if self.release_bump:
            env_params += ' -e RELEASE_BUMP=1'

        if self.verbose:
            env_params += ' -e VERBOSE=1'

        return env_params

    def preprocess(self, spec_templ):
        out_spec_basename = os.path.basename(
            spec_templ[:-5] if spec_templ.endswith('.rpkg') else spec_templ)

        out_spec_path = os.path.join(self.outdir, out_spec_basename)

        if self.preprocess_spec:
            source_params = self.get_preproc_source_params()
            env_params = self.get_preproc_env_params(spec_templ)

            cmd = 'cat {0} | preproc -C {1} {2} {3} > {4}'.format(
                quote(spec_templ), quote(self.path), env_params,
                source_params, quote(out_spec_path))

            log.debug('Running: '+cmd)

            try:
                rc = subprocess.check_call(cmd, shell=True)
            except subprocess.CalledProcessError:
                raise RpkgError('Error during spec preprocessing.')
        else:
            shutil.copy(spec_templ, out_spec_path)

        log.info('Wrote: {0}'.format(out_spec_path))
        return out_spec_path

    def load_nameverrel(self):
        """Get name, version, release of a package module."""
        cmd = ['rpm']
        cmd.extend(self.rpmdefines)

        # We make sure there is a space at the end of our query so that
        # we can split it later.  When there are subpackages, we get a
        # listing for each subpackage.  We only care about the first.
        cmd.extend(['-q', '--qf', '??%{NAME} %{EPOCH} %{VERSION} %{RELEASE}??',
                    '--specfile', '%s' % os.path.join(self.path, self.spec)])

        proc = subprocess.Popen(cmd, shell=False,
                                universal_newlines=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        output, err = proc.communicate()

        if err:
            log.debug('Errors occurred while running following command to get N-V-R-E:')
            log.debug(cmd)
            log.error(err)
        if proc.returncode > 0:
            raise RpkgError('Could not get n-v-r-e from %s'
                            % os.path.join(self.path, self.spec))

        # Get just the output, then split it by ??, grab the first and split
        # again to get ver and rel
        first_line_output = output.split('??')[1]
        parts = first_line_output.split()
        if len(parts) != 4:
            raise RpkgError('Could not get n-v-r-e from %r'
                            % first_line_output)
        (self._module_name_spec,
         self._epoch,
         self._ver,
         self._rel) = parts

        # Most packages don't include a "Epoch: 0" line, in which case RPM
        # returns '(none)'
        if self._epoch == "(none)":
            self._epoch = "0"
