import rpkglib
import os
import glob
import tempfile
import configparser
import munch
import mock
import pytest

from rpkglib.cli import rpkgClient
from rpkglib.exceptions import NotUnpackedException
from .spec_templates import SPEC_TEMPLATE
from . import base

RPKG_CONFIG = """
[rpkg]
preprocess_spec = True
auto_pack = True

[git]
lookaside = http://localhost/repo/pkgs/%(ns1)s/%(pkg)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s
lookaside_cgi = https://localhost/repo/pkgs/upload.cgi
gitbaseurl = ssh://%(user)s@localhost/%(module)s
anongiturl = git://localhost/%(module)s
"""

class TestCli(base.TestCase):
    def setUp(self):
        super(TestCli, self).setUp()
        config_fd, self.config_path = tempfile.mkstemp()
        config_file = os.fdopen(config_fd, 'w+')
        config_file.write(RPKG_CONFIG)
        config_file.close()

        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)

        self.client = rpkgClient(self.config, name='rpkg')
        self.client.do_imports('rpkglib')
        self.client.args = munch.Munch(user='user', q='q', path=self.srcdir, outdir=self.outdir, module_name=None, debug=False, v=False, target=None)
        self.client.args['with'] = []
        self.client.args['without'] = []

    def tearDown(self):
        os.unlink(self.config_path)
        super(TestCli, self).tearDown()

    @mock.patch('rpkglib.Commands')
    def test_load_cmd(self, cmds):
        self.client.load_cmd()
        cmds.assert_called_with(
            self.srcdir,
            'http://localhost/repo/pkgs/%(ns1)s/%(pkg)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s',
            'sha512',
            'https://localhost/repo/pkgs/upload.cgi',
            'ssh://%(user)s@localhost/%(module)s',
            'git://localhost/%(module)s',
            branchre='.*',
            kojiconfig='',
            build_client=None,
            user='user',
            quiet='q',
            realms=[],
            outdir=self.outdir,
            preprocess=True,
            base_output_path='/tmp/rpkg'
        )

    def test_make_source_from_packed_raises(self):
        self.make_packed_content()
        self.client.args.spec = ''
        with pytest.raises(NotUnpackedException):
            self.client.make_source()

    def test_make_source_from_unpacked(self):
        self.make_unpacked_content()
        self.client.args.spec = ''
        self.client.make_source()
        self.assertTrue(glob.glob('{0}/{1}'.format(self.outdir, '*.tar.gz')))

    def test_unpacked_changes_into_packed_by_make_sources(self):
        self.make_unpacked_content()
        self.client.args.spec = ''
        self.client.make_source()
        self.assertTrue(glob.glob('{0}/{1}'.format(self.outdir, '*.tar.gz')))
        with pytest.raises(NotUnpackedException):
            self.client.make_source()

    def test_make_srpm_from_packed(self):
        self.make_packed_content()
        self.client.args.spec = ''
        self.client.srpm()
        assert 'testpkg-1-1.src.rpm' in os.listdir(self.outdir)

    def test_make_srpm_from_unpacked(self):
        self.make_unpacked_content()
        self.client.args.spec = ''
        self.client.srpm()
        assert 'testpkg-1-1.src.rpm' in os.listdir(self.outdir)

    def test_make_srpm_multiple_specs(self):
        spec1_path = self.dump_spec(
            SPEC_TEMPLATE, pkgname='testpkg1', source0='source0.tar.gz')
        spec2_path = self.dump_spec(
            SPEC_TEMPLATE, pkgname='testpkg2', source0='source0.tar.gz')
        self.touch_file('source0.tar.gz')

        self.client.args.spec = spec1_path
        self.client.srpm()
        assert 'testpkg1-1-1.src.rpm' in os.listdir(self.outdir)

        self.client.cmd.set_specloc(spec2_path)
        self.client.cmd.load_repo_spec()
        self.client.cmd.load_spec()
        self.client.srpm()
        assert 'testpkg2-1-1.src.rpm' in os.listdir(self.outdir)

    def test_make_srpm_on_rpmtree(self):
        spec_path = self.dump_spec(
            SPEC_TEMPLATE, pkgname='testpkg1', subdir='SPECS', source0='source0.tar.gz')
        self.touch_file('source0.tar.gz', subdir='SOURCES')
        self.client.args.spec = spec_path
        path = os.path.join(self.srcdir, 'SOURCES')
        self.client.args.outdir = path
        self.client.srpm()
        assert 'testpkg1-1-1.src.rpm' in os.listdir(path)

    def test_make_srpm_on_plain_dir_no_autopack(self):
        client = rpkgClient(self.config, name='rpkg')
        client.do_imports('rpkglib')
        client.args = munch.Munch(user='user', q='q', path=self.srcdir, outdir=self.outdir, module_name=None, debug=False, v=False)
        client.auto_pack = False

        spec1_path = self.dump_spec(
            SPEC_TEMPLATE, pkgname='testpkg1', source0='source0.tar.gz')
        self.touch_file('source0.tar.gz')

        self.client.args.spec = spec1_path
        self.client.srpm()
        assert 'testpkg1-1-1.src.rpm' in os.listdir(self.outdir)
